;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Customization file for Chez Scheme.

;; Inform Soft Scheme of the types of the primitives.

;; Standard Chez Scheme primitives.
(st:defprim 'call/cc       '(((a -> b) -> a) -> a))
(st:defprim 'add1          '(num -> num)                                  'pure)
(st:defprim 'sub1          '(num -> num)                                  'pure)
(st:defprim '1+            '(num -> num)                                  'pure)
(st:defprim '1-            '(num -> num)                                  'pure)
(st:defprim 'ormap         '((a -> (+ false b)) (list a) -> (+ false b)))
(st:defprim 'andmap        '((a -> (+ true b)) (list a) -> (+ true b)))
(st:defprim 'make-list     '(num a -> (list a))                'pure-if-cons-is)
(st:defprim 'format        '(str (&rest a) -> str))
(st:defprim 'error         '(a str (&rest b) -> c)                        'pure)
(st:defprim 'printf        '(str (&rest a) -> void))
(st:defprim 'pretty-print  '(a (&optional oport) -> void))
(st:defprim 'gensym        '(-> sym)                                      'pure)
(st:defprim 'sort          '((a a -> (+ false b)) (list a) -> (list a)))
(st:defprim 'dynamic-wind  '((-> a) (-> b) (-> c) -> b))
(st:defprim 'expand-once   '(s-exp -> s-exp))
(st:defprim 'append!       '((&list (list a)) -> (list a))        'mutates-cons)
;; Rice Scheme things.
(st:defprim 'abort         '(a -> b)                                      'pure)

;; Build CHECK versions of the primitives (see checklib.scm for macros).

(defmacro CHECK-call/cc args `(CHECK-call-with-current-continuation ,@args))
(make-CHECK-typed add1 number?)
(make-CHECK-typed sub1 number?)
(make-CHECK-typed 1+ number?)
(make-CHECK-typed 1- number?)
(defmacro CHECK-ormap id
  `(lambda a
     (CHECK-INCREMENT-COUNTER ,(car id))
     (if (and (<= 2 (length a)) (procedure? (car a)) (andmap list? (cdr a)))
         (apply ormap a)
         (st:failure (cons 'CHECK-ormap ',id) "invalid arguments: ~a" a))))
(defmacro CHECK-andmap id
  `(lambda a
     (CHECK-INCREMENT-COUNTER ,(car id))
     (if (and (<= 2 (length a)) (procedure? (car a)) (andmap list? (cdr a)))
         (apply andmap a)
         (st:failure (cons 'CHECK-andmap ',id) "invalid arguments: ~a" a))))
(make-CHECK-typed make-list number? _)
(defmacro CHECK-format id
  `(lambda a
     (CHECK-INCREMENT-COUNTER ,(car id))
     (if (and (pair? a) (string? (car a)))
         (apply format a)
         (st:failure (cons 'CHECK-format ',id) "invalid arguments: ~a" a))))
(defmacro CHECK-error id
  `(lambda a
     (CHECK-INCREMENT-COUNTER ,(car id))
     (if (and (pair? a) (pair? (cdr a)) (string? (cadr a)))
         (apply error a)
         (st:failure (cons 'CHECK-error ',id) "invalid arguments: ~a" a))))
(defmacro CHECK-printf id
  `(lambda a
     (CHECK-INCREMENT-COUNTER ,(car id))
     (if (and (pair? a) (string? (car a)))
         (apply printf a)
         (st:failure (cons 'CHECK-printf ',id) "invalid arguments: ~a" a))))
(defmacro CHECK-pretty-print id
  `(lambda a
     (CHECK-INCREMENT-COUNTER ,(car id))
     (if (and (pair? a) (or (null? (cdr a)) (output-port? (cadr a))))
         (apply pretty-print a)
         (st:failure (cons 'CHECK-pretty-print ',id) "invalid arguments: ~a" a))))
(make-CHECK-typed gensym)
(make-CHECK-typed sort procedure? list?)
(make-CHECK-typed dynamic-wind procedure? procedure? procedure?)
(make-CHECK-typed expand-once _)
(defmacro CHECK-append! id
  `(lambda a
     (CHECK-INCREMENT-COUNTER ,(car id))
     (let loop ((b a))
       (match b
         (() #t)
         ((l) #t)
         (((? list?) . y) (loop y))
         (_ (st:failure (cons 'CHECK-append! ',id) "invalid arguments: ~a" a))))
     (apply append! a)))
(make-CHECK-typed abort _)
