;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;; requires expand-once, error, and gensym

(defmacro encapsulate args
  (letrec
    ((expand-to-defines
      (lambda (def)
        (match def
          (('define (? symbol? n) e)
           (list def))
          (('define ((? symbol? n) . args) . body)
           (list `(define ,n (lambda ,args ,@body))))
          (('begin . (? list? defs))
           (foldr append '() (map expand-to-defines defs)))
          (((? symbol? k) . _)
           (=> fail)
           (let ((d2 (expand-once def)))
             (if (equal? def d2)
                 (fail)
                 (expand-to-defines d2))))
          (_ (list `(define ,(gensym) ,def))))))
     (readfile
      (lambda (f)
        (with-input-from-file f
          (letrec ((rf (lambda ()
                         (match (read)
                           ((? eof-object?) '())
                           (sexp (cons sexp (rf)))))))
            rf))))
     (foldr
      (lambda (f i l)
        (recur loop ((l l))
          (match l
            (() i)
            ((x . y) (f x (loop y))))))))
    (match args
      ((((? symbol? exports) ...) (files ...) init ...)
       (let* ((do-init (gensym))
              (exports (cons do-init exports))
              (pgm `(begin ,@(map (lambda (f) `(begin ,@(readfile f))) (map eval files)) (define ,do-init (lambda () ,@init #f))))
              (defs (expand-to-defines pgm)))
         `(begin
            (match-define ,exports
              (let ,(map (match-lambda (('define x _) `(,x (void))))
                         defs)
                ,@(map (match-lambda (('define x e) `(set! ,x ,e)))
                       defs)
                (list ,@exports)))
            (,do-init))))
      (_ (error 'encapsulate "invalid syntax")))))
