;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type Inference

(define GenLet #t)
(define GenMatch #t)
(define LetOnce #f)

(define type-defs
  (lambda (d)
    (for-each
      (match-lambda
        ((? Defstruct? b) (type-structure b))
        ((? Datatype? b) (type-structure b))
        (c (type-component c #t)))
      (make-components d))
    (close '())))
      
(define type-structure
  (match-lambda
    (($ Defstruct x _ make pred get set getn setn mutable)
     (let* ((vars
             (map (lambda (_) (gensym)) get))
            (make-get-type
             (lambda (getter v)
               (match getter
                 (($ Some b)
                  (set-Name-ty! b
                    (closeall (R+ initial-type-env `((,x ,@vars) -> ,v)))))
                 (_ #f))))
            (make-set-type
             (lambda (setter v)
               (match setter
                 (($ Some b)
                  (set-Name-ty! b
                    (closeall (R+ initial-type-env `((,x ,@vars) ,v -> void)))))
                 (_ #f)))))
       (set-Name-ty! make
         (closeall (R+ initial-type-env `(,@vars -> (,x ,@vars)))))
       (set-Name-ty! pred
         (closeall (R+ initial-type-env `((+ (,x ,@vars) y) -> bool))))
       (for-each2 make-get-type get vars)
       (for-each2 make-set-type set vars)
       (for-each2 make-get-type getn vars)
       (for-each2 make-set-type setn vars)))
    (($ Datatype dt)
     (for-each
       (match-lambda
         ((type . variants)
          (for-each
            (match-lambda
              (($ Variant con pred arg-types)
               (set-Name-ty! con
                 (closeall (R+ initial-type-env
                               `(,@(cdr arg-types) -> ,type))))
               (set-Name-ty! pred
                 (closeall (R+ initial-type-env
                               `((+ ,(Name-predicate pred) x) -> bool))))))
            variants)))
       dt))))

(define type-component
  (lambda (component top)
    (when verbose
      (let ((cnames (filter-map
                      (match-lambda (($ Define b _) (Name-name b)))
                      component)))
        (unless (null? cnames)
          (printf "Typing ~a~%" cnames))))
    (let* ((f (match-lambda (($ Define b e) (make-Bind b e))))
           (bindings (map f component))
           (names (map (match-lambda (($ Define b _) (pname b))) component))
           (f1 (match-lambda
                 (($ Define b _)
                  (set-Name-ty! b (Tvar)))))
           (f2 (match-lambda
                 ((and d ($ Define b e))
                  (set-Define-exp! d (W e names)))))
           (f3 (match-lambda
                 (($ Define b e)
                  (unify (Name-ty b) (typeof e)))))
           (f4 (match-lambda
                 (($ Define b _)
                  (Name-ty b))))
           (f5 (lambda (d ts)
                 (match d
                   (($ Define b _)
                    (set-Name-ty! b ts))))))
      (push-level)
      (for-each f1 component)
      (for-each f2 component)
      (for-each f3 component)
      (for-each limit-expansive component)
      (for-each f5 component (close (map f4 component)))
      (pop-level))))

(define W
  (lambda (e component)
    (match e
      (($ Const _ pred)
       (make-Type (R+ initial-type-env (Name-predicate pred)) e))
      (($ Var x)
       (unless (Name-ty x)
         ; set up type for unbound references
         (set-Name-ty! x
           (if (Name-mutated x)
               (monoTvar)
               (let* ((_1 (push-level))
                      (t (closeall (Tvar)))
                      (_2 (pop-level)))
                 t))))
       (if (TS? (Name-ty x))
           (match-let* ((tynode (make-Type #f #f))
                        ((t absv) (instantiate (Name-ty x) tynode)))
             (set-Type-ty! tynode t)
             (set-Type-exp! tynode
               (match (Name-primitive x)
                 ('imprecise
                   (make-Check (list absv #f #f #f component) e))
                 ('check
                   (make-Check (list (cons Top absv) #f #f #f component) e))
                 ('nocheck
                   e)
                 (#t
                   (make-Check (list absv (mk-definite-prim t) #f #f component) e))
                 (#f
                   (make-Check (list absv #f #f #t component) e))))
             tynode)
           e))
      (($ Lam x e1)
       (for-each (lambda (b) (set-Name-ty! b (Tvar))) x)
       (match-let* ((body (W e1 component))
                    ((t absv) (R+collect initial-type-env
                                `(,@(map Name-ty x) -> ,(typeof body)))))
         (make-Type t
           (make-Check (list absv (mk-definite-lam t) #f #f component)
             (make-Lam x body)))))
      (($ Vlam x rest e1)
       (for-each (lambda (b) (set-Name-ty! b (Tvar))) x)
       (match-let* ((z (Tvar))
                    (_ (set-Name-ty! rest (R+ initial-type-env `(list ,z))))
                    (body (W e1 component))
                    ((t absv) (R+collect initial-type-env
                                `(,@(map Name-ty x) (&list ,z) -> ,(typeof body)))))
         (make-Type t
           (make-Check (list absv (mk-definite-lam t) #f #f component)
             (make-Vlam x rest body)))))
      (($ App e0 args)
       (match-let* ((t0 (W e0 component))
                    (targs (mapLR (lambda (e) (W e component)) args))
                    (a* (map (lambda (_) (Tvar)) args))
                    (b (Tvar))
                    ((t absv)
                     (R-collect initial-type-env `(,@a* -> ,b)))
                    (definf (mk-definite-app t)))
         (unify (typeof t0) t)
         (for-each2 unify (map typeof targs) a*)
         (if (syntactically-a-procedure? t0)
             (make-Type b
               (make-App t0 targs))
             (make-Type b
               (make-Check (list absv definf #f #f component)
                 (make-App t0 targs))))))
      (($ Let b e2)
       (let* ((do-bind
               (match-lambda
                 (($ Bind b e)
                  (if GenLet
                      (let* ((_ (push-level))
                             (e (W e (list (pname b))))
                             (bind (make-Bind b e)))
                        (limit-expansive bind)
                        (set-Name-ty! b (car (close (list (typeof e)))))
                        (pop-level)
                        bind)
                      (let ((e (W e component)))
                        (set-Name-ty! b (typeof e))
                        (make-Bind b e))))))
              (tb (map do-bind b))
              (body (W e2 component)))
         (make-Let tb body)))
      (($ Let* b e2)
       (let* ((do-bind
               (match-lambda
                 (($ Bind b e)
                  (if GenLet
                      (let* ((_ (push-level))
                             (e (W e (list (pname b))))
                             (bind (make-Bind b e)))
                        (limit-expansive bind)
                        (set-Name-ty! b (car (close (list (typeof e)))))
                        (pop-level)
                        bind)
                      (let ((e (W e component)))
                        (set-Name-ty! b (typeof e))
                        (make-Bind b e))))))
              (tb (mapLR do-bind b))
              (body (W e2 component)))
         (make-Let* tb body)))
      (($ Letr b e2)
       ;; this code reorders letrecs, which is ok by R4RS
       (let* ((do-comp
               (lambda (b)
                 (if GenLet
                     (let* ((f1 (match-lambda
                                  (($ Bind b _) (set-Name-ty! b (Tvar)))))
                            (names (map (match-lambda
                                          (($ Bind b _) (pname b)))
                                        b))
                            (f2 (match-lambda
                                  (($ Bind b e) (make-Bind b (W e names)))))
                            (f3 (match-lambda
                                  (($ Bind b e) (unify (Name-ty b) (typeof e))
                                   (Name-ty b))))
                            (f4 (lambda (bind ts)
                                  (match bind
                                    (($ Bind b _) (set-Name-ty! b ts)))))
                            (_1 (push-level))
                            (_2 (for-each f1 b))
                            (tb (mapLR f2 b))
                            (_3 (for-each limit-expansive tb))
                            (ts-list (close (mapLR f3 tb))))
                       (pop-level)
                       (for-each2 f4 tb ts-list)
                       tb)
                     (let* ((f1 (match-lambda
                                  (($ Bind b _) (set-Name-ty! b (Tvar)))))
                            (f2 (match-lambda
                                  (($ Bind b e) (make-Bind b (W e component)))))
                            (f3 (match-lambda
                                  (($ Bind b e) (unify (Name-ty b) (typeof e)))))
                            (_1 (for-each f1 b))
                            (tb (mapLR f2 b)))
                       (for-each f3 tb)
                       tb))))
              (comps (make-letrec-components b))
              (tb (foldr append '() (mapLR do-comp comps))))
         (make-Letr tb (W e2 component))))
      (($ Body defs exps)
       (for-each
         (match-lambda
           ((? Defstruct? b) (type-structure b))
           ((? Datatype? b) (type-structure b))
           (c (type-component c #f)))
         (make-body-components defs))
       (let ((texps (mapLR (lambda (x) (W x component)) exps)))
         (make-Body defs texps)))
      (($ And exps)
       (let* ((texps (mapLR (lambda (x) (W x component)) exps))
              (t (match texps
                   (() (R+ initial-type-env 'true))
                   ((e) (typeof e))
                   (_ (let ((a (R+ initial-type-env 'false)))
                        (unify (typeof (rac texps)) a)
                        a)))))
         (make-Type t
           (make-And texps))))
      (($ Or exps)
       (let* ((texps (mapLR (lambda (x) (W x component)) exps))
              (t (match texps
                   (() (R+ initial-type-env 'false))
                   ((e) (typeof e))
                   (_ (let* ((t-last (typeof (rac texps)))
                             (but-last (rdc texps))
                             (a (Tvar)))
                        (for-each
                          (lambda (e)
                            (unify
                              (typeof e)
                              (R+ initial-type-env `(+ (not false) ,a))))
                          but-last)
                        (unify
                          t-last
                          (R+ initial-type-env `(+ (not false) ,a)))
                        t-last)))))
         (make-Type t
           (make-Or texps))))
      (($ Begin exps)
       (let ((texps (mapLR (lambda (x) (W x component)) exps)))
         (make-Begin texps)))
      (($ If test then els)
       (let ((ttest (W test component))
             (tthen (W then component))
             (tels (W els component))
             (a (Tvar)))
         (unify (typeof tthen) a)
         (unify (typeof tels) a)
         (make-Type a
           (make-If ttest tthen tels))))
      (($ Delay e2)
       (let ((texp (W e2 component)))
         (make-Type (R+ initial-type-env `(promise ,(typeof texp)))
           (make-Delay texp))))
      (($ Set! x body)
       (unless (Name-ty x) ; set up type for unbound references
         (set-Name-ty! x (monoTvar)))
       (let* ((body (W body component))
              (t (if (TS? (Name-ty x))
                     (car (instantiate (Name-ty x) #f))
                     (Name-ty x))))
         (unify t (typeof body))
         (make-Type (R+ initial-type-env 'void)
           (make-Set! x body))))
      (($ Record bind)
       (let* ((tbind (map (match-lambda
                            (($ Bind name exp)
                             (make-Bind name (W exp component))))
                          bind))
              (t (R+ initial-type-env
                     `(record ,@(map (match-lambda
                                       (($ Bind name exp)
                                        (list name (typeof exp))))
                                     tbind)))))
         (make-Type t
           (make-Record tbind))))
      (($ Field name exp)
       (match-let* ((texp (W exp component))
                    (a (Tvar))
                    ((t absv) (R-collect initial-type-env
                                `(record (,name ,a)))))
         (unify (typeof texp) t)
         (make-Type a
           (make-Check (list absv #f #f #f component)
             (make-Field name texp)))))
      (($ Cast (ty t absv) exp)
       (let ((texp (W exp component))
             (a (Tvar)))
         (unify (R+ initial-type-env `(,(typeof texp) -> ,a)) t)
         (make-Type a
           (make-Check (list absv #f #f #f component)
             (make-Cast (list ty t absv) texp)))))
      (($ Match exp clauses)
       (for-each
         (match-lambda
           (($ Mclause p _ (? Name? fail))
            (set-Name-ty! fail (R+ initial-type-env '(a ?-> b))))
           (_ #f))
         clauses)
       (match-let* ((iclauses (improve-clauses
                                (append clauses
                                  (list (make-Mclause (make-Pelse) #f #f)))))
                    ((tmatch absv precise)
                     (W-match (rdc iclauses) (rac iclauses)))
                    (texp (W exp component))
                    (_ (unify (typeof texp) tmatch))
                    (tclauses (mapLR
                                (match-lambda
                                  (($ Mclause p e fail)
                                   (make-Mclause p (W e component) fail)))
                                clauses))
                    (a (Tvar)))
         (for-each
           (match-lambda (($ Mclause _ e _) (unify (typeof e) a)))
           tclauses)
         (make-Type a
           (make-Check (list absv #f (not precise) #f component)
             (make-Match texp tclauses))))))))

(define W-match
  (lambda (clauses last)
    (letrec
      ((bindings '())
       (encode
        (match-lambda
          (($ Pand pats) (encode* pats))
          (x             (encode* (list x)))))
       (encode*
        (lambda (pats)
          (let* ((concrete? (lambda (p)
                              (or (Pconst? p) (Pobj? p) (Ppred? p) (Pelse? p))))
                 (var? (lambda (p) (or (Pvar? p) (Pany? p))))
                 (not-var? (lambda (p) (and (not (Pvar? p)) (not (Pany? p)))))
                 (t (match (filter concrete? pats)
                      ((p) (R+ initial-type-env
                               (match (template p)
                                 ((x) x)
                                 (x `(+ ,@x)))))
                      (()  (R+ initial-type-env
                               `(+ ,@(apply append
                                       (map template (filter not-var? pats)))
                                   ,@(if (null? (filter var? pats))
                                         '()
                                         (list (out1Tvar)))))))))
            (for-each
              (match-lambda
                (($ Pvar b)
                 (set! bindings (cons b bindings))
                 (set-Name-ty! b (pat-var-bind t))))
              (filter Pvar? pats))
            t)))
       (template
        (match-lambda
          ((? Pelse?) '())
          (($ Pconst _ pred) (list (Name-predicate pred)))
          ((and pat ($ Pobj c args))
           (list
             (cond ((or (eq? %vector? c) (eq? %cvector? c))
                    (cons (if (eq? %vector? c)
                              'vec
                              'cvec)
                          (match (mapLR encode args)
                            (() (list (out1Tvar)))
                            ((first . rest)
                             (list (foldr (lambda (x y) (unify x y) y)
                                     first
                                     rest))))))
                   (else
                    (cons (car (Name-predicate c))
                          (mapLR encode args))))))
          (($ Ppred pred)
           (cond ((eq? pred %boolean?) (list 'true 'false))
                 ((eq? pred %list?) (list `(list ,(out1Tvar))))
                 (else (list (cons (car (Name-predicate pred))
                                   (mapLR
                                     (lambda (_) (out1Tvar))
                                     (cdr (Name-predicate pred))))))))
          (($ Pnot (? Pconst?)) '())
          (($ Pnot ($ Ppred pred))
           (cond ((eq? pred %boolean?) '((not true) (not false)))
                 ((eq? pred %procedure?) '((not ?->)))
                 ((eq? pred %list?) '())
                 (else `((not ,(car (Name-predicate pred)))))))
          (($ Pnot ($ Pobj pred pats))
           (let ((m (foldr + 0 (map non-triv pats))))
             (case m
               ((0) 
                `((not ,(car (Name-predicate pred)))))
               ((1)
                `((,(car (Name-predicate pred))
                   ,@(map (match-lambda
                            (($ Pobj pred _)
                             `(+ (not ,(car (Name-predicate pred))) ,(out1Tvar)))
                            (($ Ppred pred)
                             `(+ (not ,(car (Name-predicate pred))) ,(out1Tvar)))
                            (_ (out1Tvar)))
                          pats))))
               (else
                '()))))))
       (non-triv
        (match-lambda
          ((? Pvar?) 0)
          ((? Pany?) 0)
          ((? Pelse?) 0)
          ((? Pconst?) 2)
          (($ Pobj _ pats) (foldr + 1 (map non-triv pats)))
          (_ 1)))
       (precise
        (match-lambda
          ((? Pconst?) #f)
          (($ Pand pats) (andmap precise pats))
          (($ Pnot pat) (precise pat))
          (($ Pobj pred pats)
           (let ((m (foldr + 0 (map non-triv pats))))
             (case m
               ((0) #t)
               ((1) (andmap precise pats))
               (else #f))))
          (($ Ppred pred)
           (not (eq? pred %list?)))
          (_ #t))))
      (push-level)
      (match-let* ((precise-match
                    (and (andmap
                           (match-lambda (($ Mclause _ _ fail) (not fail)))
                           clauses)
                         (match last
                           (($ Mclause p _ _) (precise p)))))
                   (types (mapLR
                            (match-lambda (($ Mclause p _ _) (encode p)))
                            clauses))
                   ((t absv)
                    (R-match (foldr (lambda (x y) (unify x y) y)
                               (Tvar)
                               types))))
        (unify (out1Tvar) t)
        (for-each limit-name bindings)
        (for-each2 set-Name-ty! bindings (close (map Name-ty bindings)))
        (pop-level)
        '(pretty-print `(MATCH-input
                          ,@(map (match-lambda
                                   (($ Mclause p _ _) (ppat p)))
                                 clauses)))
        '(pretty-print `(MATCH-type
                          ,(ptype t)
                          ,@(map (lambda (b)
                                   (list (pname b) (ptype (Name-ty b))))
                                 bindings)))
        (list t absv precise-match)))))

(define syntactically-a-procedure?
  (match-lambda
    (($ Type _ e) (syntactically-a-procedure? e))
    (($ Check _ e) (syntactically-a-procedure? e))
    (($ Var x) (Name-primitive x))
    ((? Lam?) #t)
    ((? Vlam?) #t)
    (($ Let _ body)  (syntactically-a-procedure? body))
    (($ Let* _ body) (syntactically-a-procedure? body))
    (($ Letr _ body) (syntactically-a-procedure? body))
    (($ If _ e2 e3) (and (syntactically-a-procedure? e2)
                         (syntactically-a-procedure? e3)))
    (($ Begin exps) (syntactically-a-procedure? (rac exps)))
    (($ Body _ exps) (syntactically-a-procedure? (rac exps)))
    (_ #f)))
      
(define typeof
  (match-lambda
    (($ Type t _) t)
    (($ Check _ e) (typeof e))
    (($ Let _ body) (typeof body))
    (($ Let* _ body) (typeof body))
    (($ Letr _ body) (typeof body))
    (($ Body _ exps) (typeof (rac exps)))
    (($ Begin exps) (typeof (rac exps)))
    (($ Var x) (Name-ty x))))

(define limit-name
  (lambda (n)
    (when (Name-mutated n)
      (unify (Name-ty n) (out1Tvar)))))

(define limit-expansive
  (letrec ((limit!
            (lambda (t)
              (unify t (out1Tvar))))
           (expansive-pattern?
            (match-lambda
              ((? Pconst?) #f)
              (($ Pvar x) (Name-mutated x))
              (($ Pobj _ pats) (ormap expansive-pattern? pats))
              ((? Pany?) #f)
              ((? Pelse?) #f)
              (($ Pand pats) (ormap expansive-pattern? pats))
              (($ Ppred x) (Name-mutated x))
              (($ Pnot pat) (expansive-pattern? pat))))
           (limit-expr
            (match-lambda
              (($ Bind b e)
               (if (Name-mutated b)
                   (limit! (typeof e))
                   (limit-expr e)))
              ((? Defstruct?) #f)
              ((? Datatype?) #f)
              (($ Define x e)
               (if (and x (Name-mutated x))
                   (limit! (typeof e))
                   (limit-expr e)))
              (($ Type t ($ App ($ Type _ ($ Check _ ($ Var x))) exps))
               (cond ((list? (Name-pure x))
                      (if (= (length (Name-pure x)) (length exps))
                          (for-each2 (lambda (pure e)
                                       (if pure
                                           (limit-expr e)
                                           (limit! (typeof e))))
                            (Name-pure x)
                            exps)
                          (limit! t)))
                     ((or (eq? #t (Name-pure x))
                          (and (eq? 'cons (Name-pure x)) (not cons-is-mutable)))
                      (for-each limit-expr exps))
                     (else
                      (limit! t))))
              (($ Type t ($ App _ _)) (limit! t))
              (($ Type t ($ Check _ ($ App _ _))) (limit! t))
              (($ Delay _) #f)
              (($ Type t ($ Set! _ _)) (limit! t))
              (($ Var _) #f)
              ((? Const?) #f)
              (($ Lam _ _) #f)
              (($ Vlam _ _ _) #f)
              (($ Let bind body)
               (limit-expr body)
               (for-each limit-expr bind))
              (($ Let* bind body)
               (limit-expr body)
               (for-each limit-expr bind))
              (($ Letr bind body)
               (limit-expr body)
               (for-each limit-expr bind))
              (($ Body defs exps)
               (for-each limit-expr defs)
               (for-each limit-expr exps))
              (($ And exps) (for-each limit-expr exps))
              (($ Or exps) (for-each limit-expr exps))
              (($ Begin exps) (for-each limit-expr exps))
              (($ If e1 e2 e3)
               (limit-expr e1)
               (limit-expr e2)
               (limit-expr e3))
              (($ Record bind)
               (for-each (match-lambda (($ Bind _ e) (limit-expr e))) bind))
              (($ Field _ exp) (limit-expr exp))
              (($ Cast _ exp) (limit-expr exp))
              (($ Match exp clauses)
               (limit-expr exp)
               (for-each
                 (match-lambda (($ Mclause pat body fail)
                                (if (or (and fail (Name-mutated fail))
                                        (expansive-pattern? pat))
                                    (limit! (typeof body))
                                    (limit-expr body))))
                 clauses))
              (($ Type _ e1) (limit-expr e1))
              (($ Check _ e1) (limit-expr e1)))))
    limit-expr))
